@@ -1,38 +1,66 @@
 # Implementation based on tf.keras.engine.training.py
 # https://github.com/tensorflow/tensorflow/blob/v2.2.0/tensorflow/python/keras/engine/training.py
 
-import io
 import pickle
+import threading
 import typing as tp
+from contextlib import contextmanager
 from copy import copy
-from io import StringIO
+from dataclasses import dataclass
 from pathlib import Path
 
 import jax
 import jax.numpy as jnp
 import numpy as np
-import rich
-import rich.box
-import toolz
-import yaml
-from elegy import types, utils
-from elegy.callbacks import Callback, CallbackList, History
-from elegy.data import (
-    DataHandler,
-    map_append,
-    map_structure,
-    train_validation_split,
-    unpack_x_y_sample_weight,
-)
+import treex as tx
+
+from elegy import data, types, utils
+from elegy.callbacks import Callback, CallbackList, History, history
+from elegy.callbacks.sigint import SigIntMode
 from elegy.data import utils as data_utils
-from elegy.model.model_core import ModelCore, PredStep, TestStep
-from rich.table import Table
-from rich.text import Text
+from elegy.model.model_core import ModelCore, PredStepOutput, TestStepOutput
 
-# from elegy.module import Module
+__all__ = ["ModelBase", "load"]
 
 
-__all__ = ["Model", "load"]
+@dataclass
+class _ModelContext(threading.local):
+    model: tp.Optional["ModelBase"] = None
+    history: tp.Optional[History] = None
+    stop_training: bool = False
+
+    def __enter__(self):
+        global _MODEL_CONTEXT
+        self._old_context = _MODEL_CONTEXT
+        _MODEL_CONTEXT = self
+
+    def __exit__(self, *args):
+        global _MODEL_CONTEXT
+        _MODEL_CONTEXT = self._old_context
+
+    @contextmanager
+    def update(self, **kwargs):
+        fields = vars(self).copy()
+        fields.pop("_old_context", None)
+        fields.update(kwargs)
+
+        with _ModelContext(**fields):
+            yield
+
+    @contextmanager
+    def callbacks_context(self, model: tp.Optional["ModelCore"]):
+        if model is None:
+            yield
+        else:
+            with _MODEL_CONTEXT.update(
+                model=model,
+                history=None,
+                stop_training=False,
+            ):
+                yield
+
+
+_MODEL_CONTEXT = _ModelContext()
 
 
 class ModelBase(ModelCore):
@@ -47,7 +75,7 @@ class ModelBase(ModelCore):
     >>> import jax.numpy as jnp
 
     >>> class MLP(elegy.Module):
-    ...     def call(self, x: jnp.ndarray) -> jnp.ndarray:
+    ...     def __call__(self, x: jnp.ndarray) -> jnp.ndarray:
     ...         x = elegy.nn.Flatten()(x)
     ...         x = elegy.nn.Linear(5)(x)
     ...         x = jax.nn.relu(x)
@@ -69,10 +97,10 @@ class ModelBase(ModelCore):
     model = elegy.Model(
         module=MLP(),
         loss=[
-            elegy.losses.SparseCategoricalCrossentropy(from_logits=True),
-            elegy.regularizers.GlobalL2(l=1e-5),
+            elegy.losses.Crossentropy(),
+            elegy.regularizers.L2(l=1e-5),
         ],
-        metrics=elegy.metrics.SparseCategoricalAccuracy(),
+        metrics=elegy.metrics.Accuracy(),
         optimizer=optax.rmsprop(1e-3),
     )
     ```
@@ -88,14 +116,14 @@ class ModelBase(ModelCore):
     ```python
     model = elegy.Model(
         module=MLP(n1=3, n2=1),
-        loss=elegy.losses.SparseCategoricalCrossentropy(from_logits=True),
-        metrics=elegy.metrics.SparseCategoricalAccuracy(),
+        loss=elegy.losses.Crossentropy(),
+        metrics=elegy.metrics.Accuracy(),
         optimizer=elegy.Optimizer(
             optax.adam(1.0), # <---- important to set this to 1.0
             lr_schedule=lambda step, epoch: 1 / (epoch * 100 + step),
             steps_per_epoch=1000,
         ),
-        run_eagerly=True,
+        eager=True,
     )
 
     history = model.fit(
@@ -109,6 +137,34 @@ class ModelBase(ModelCore):
     at the end.
     """
 
+    @property
+    def history(self) -> tp.Optional[History]:
+        if _MODEL_CONTEXT.model is not self:
+            raise ValueError(f"Trying to get `.history` outside `callbacks_context`")
+
+        return _MODEL_CONTEXT.history
+
+    @history.setter
+    def history(self, value: History):
+        if _MODEL_CONTEXT.model is not self:
+            raise ValueError(f"Trying to set `.history` outside `callbacks_context`")
+
+        _MODEL_CONTEXT.history = value
+
+    @property
+    def stop_training(self) -> bool:
+        if _MODEL_CONTEXT.model is not self:
+            raise ValueError(f"Trying to get `.history` outside `callbacks_context`")
+
+        return _MODEL_CONTEXT.stop_training
+
+    @stop_training.setter
+    def stop_training(self, value: bool):
+        if _MODEL_CONTEXT.model is not self:
+            raise ValueError(f"Trying to set `.history` outside `callbacks_context`")
+
+        _MODEL_CONTEXT.stop_training = value
+
     __all__ = [
         "evaluate",
         "fit",
@@ -126,205 +182,17 @@ class ModelBase(ModelCore):
         "states",
     ]
 
-    def init(
-        self,
-        x: tp.Optional[tp.Any] = None,
-        y: tp.Optional[tp.Any] = None,
-        batch_size: tp.Optional[int] = None,
-        class_weight: tp.Optional[tp.Mapping[str, float]] = None,
-        sample_weight: tp.Optional[np.ndarray] = None,
-    ) -> None:
-
-        if self.initialized:
-            return
-
-        if x is None:
-            x = {}
-
-        data_handler = DataHandler(
-            x=x,
-            y=y,
-            sample_weight=sample_weight,
-            batch_size=batch_size,
-            class_weight=class_weight,
-            steps_per_epoch=1,
-        )
-
-        epoch, iterator = next(data_handler.enumerate_epochs())
-        batch = next(iterator)
-        x_batch, y_batch, sample_weight = unpack_x_y_sample_weight(batch)
-
-        self.init_on_batch(
-            x=x_batch,
-            y_true=y_batch,
-            sample_weight=sample_weight,
-            class_weight=class_weight,
-        )
-
-    def summary(
-        self,
-        x: tp.Optional[tp.Any] = None,
-        depth: int = 2,
-        tablefmt: str = "fancy_grid",
-        return_repr: bool = False,
-        initialize: bool = False,
-        eval_shape: bool = True,
-        **tablulate_kwargs,
-    ) -> tp.Optional[str]:
-        """
-        Prints a summary of the network.
-        Arguments:
-            x: A sample of inputs to the network.
-            depth: The level number of nested level which will be showed.
-                Information about summaries from modules deeper than `depth`
-                will be aggregated together.
-            tablefmt: A string representing the style of the table generated by
-                `tabulate`. See
-                [python-tabulate](https://github.com/astanin/python-tabulate)
-                for more options.
-            tablulate_kwargs: Additional keyword arguments passed to `tabulate`.
-                See [python-tabulate](https://github.com/astanin/python-tabulate)
-                for more options.
-        """
-
-        if x is None:
-            x = {}
-
-        entries: tp.List[types.SummaryTableEntry]
-        states = self.states.copy() if self.run_eagerly else self.states
-
-        method = (
-            self.call_summary_step if self.run_eagerly else self.call_summary_step_jit
-        )
-
-        if eval_shape:
-            entries = jax.eval_shape(self.call_summary_step, x, states)
-        else:
-            entries = method(x, states)
-
-        total_entry = entries[-1]
-        entries = entries[:-1]
-
-        depth_groups: tp.Dict[str, tp.List[types.SummaryTableEntry]] = toolz.groupby(
-            lambda entry: "/".join(entry.path.split("/")[:depth]), entries
-        )
-
-        entries = [
-            utils.get_grouped_entry(entry, depth_groups)
-            for entry in entries
-            if entry.path in depth_groups
-        ]
-
-        main_table = Table(
-            show_header=True,
-            show_lines=True,
-            show_footer=True,
-            # box=rich.box.HORIZONTALS,
-        )
-
-        main_table.add_column("Layer")
-        main_table.add_column("Outputs Shape")
-        main_table.add_column("Trainable\nParameters")
-        main_table.add_column("Non-trainable\nParameters")
-
-        rows: tp.List[tp.List[str]] = []
-
-        rows.append(["Inputs", utils.format_output(x), "", ""])
-
-        for entry in entries:
-            rows.append(
-                [
-                    f"{entry.path}{{pad}}  "
-                    + (
-                        f"[dim]{entry.module_type_name}[/]"
-                        if entry.module_type_name
-                        else ""
-                    ),
-                    utils.format_output(entry.output_value),
-                    f"[green]{entry.trainable_params_count:,}[/]{{pad}}    {utils.format_size(entry.trainable_params_size)}"
-                    if entry.trainable_params_count > 0
-                    else "",
-                    f"[green]{entry.non_trainable_params_count:,}[/]{{pad}}    {utils.format_size(entry.non_trainable_params_size)}"
-                    if entry.non_trainable_params_count > 0
-                    else "",
-                ]
-            )
-
-        # global summaries
-        params_count = total_entry.trainable_params_count
-        params_size = total_entry.trainable_params_size
-        states_count = total_entry.non_trainable_params_count
-        states_size = total_entry.non_trainable_params_size
-        total_count = params_count + states_count
-        total_size = params_size + states_size
-
-        rows.append(
-            [
-                "",
-                "Total",
-                (
-                    f"[green]{params_count:,}[/]{{pad}}    {utils.format_size(params_size)}"
-                    if params_count > 0
-                    else ""
-                ),
-                (
-                    f"[green]{states_count:,}[/]{{pad}}    {utils.format_size(states_size)}"
-                    if states_count > 0
-                    else ""
-                ),
-            ]
-        )
-
-        # add padding
-        for col in range(4):
-            max_length = max(
-                len(line.split("{pad}")[0])
-                for row in rows
-                for line in row[col].split("\n")
-            )
-
-            for row in rows:
-                row[col] = "\n".join(
-                    line.format(
-                        pad=" " * (max_length - len(line.rstrip().split("{pad}")[0]))
-                    )
-                    for line in row[col].rstrip().split("\n")
-                )
-
-        for row in rows[:-1]:
-            main_table.add_row(*row)
-
-        main_table.columns[1].footer = Text.from_markup(rows[-1][1], justify="right")
-        main_table.columns[2].footer = rows[-1][2]
-        main_table.columns[3].footer = rows[-1][3]
-        main_table.caption_style = "bold"
-        main_table.caption = (
-            "\nTotal Parameters: "
-            + f"[green]{total_count:,}[/]   {utils.format_size(total_size)}"
-            if total_count > 0
-            else ""
-        )
-
-        summary = "\n" + utils.get_table_repr(main_table)
-
-        print(summary)
-
-        if return_repr:
-            return summary
-
     def fit(
         self,
-        x: tp.Optional[tp.Any] = None,
-        y: tp.Optional[tp.Any] = None,
-        batch_size: tp.Optional[int] = None,
+        inputs: tp.Optional[tp.Any] = None,
+        labels: tp.Optional[tp.Any] = None,
+        batch_size: tp.Optional[int] = 32,
         epochs: int = 1,
         verbose: int = 1,
         callbacks: tp.Union[tp.List[Callback], CallbackList, None] = None,
         validation_split: float = 0.0,
         validation_data: tp.Union[tp.Tuple, tp.Iterable, None] = None,
         shuffle: bool = True,
-        class_weight: tp.Optional[tp.Mapping[str, float]] = None,
-        sample_weight: tp.Optional[np.ndarray] = None,
         initial_epoch: int = 0,
         steps_per_epoch: tp.Optional[int] = None,
         validation_steps: tp.Optional[int] = None,
@@ -482,129 +350,135 @@ class ModelBase(ModelCore):
             ValueError: In case of mismatch between the provided input data
                 and what the model expects.
         """
-        if x is None:
-            x = {}
 
-        if not self.initialized:
-            self.init(
-                x=x,
-                y=y,
-                batch_size=batch_size,
-                class_weight=class_weight,
-                sample_weight=sample_weight,
-            )
+        if batch_size is not None:
+            batch_size = self._distributed_strategy.lift_batch_size(batch_size)
 
-        if validation_split:
-            # Create the validation data using the training data. Only supported for
-            # `Jax Numpy` and `NumPy` input.
-            (x, y, sample_weight), validation_data = train_validation_split(
-                (x, y, sample_weight), validation_split=validation_split, shuffle=False
-            )
+        with _MODEL_CONTEXT.callbacks_context(self):
+            if inputs is None:
+                inputs = dict()
 
-        self.stop_training = False
-        data_handler = DataHandler(
-            x=x,
-            y=y,
-            sample_weight=sample_weight,
-            batch_size=batch_size,
-            steps_per_epoch=steps_per_epoch,
-            initial_epoch=initial_epoch,
-            epochs=epochs,
-            shuffle=shuffle,
-            class_weight=class_weight,
-        )
-        # Container that configures and calls `tf.keras.Callback`s.
-        if not isinstance(callbacks, CallbackList):
-            callbacks = CallbackList(
-                callbacks,
-                add_history=True,
-                add_progbar=verbose != 0,
-                model=self,
-                verbose=verbose,
-                epochs=epochs,
-                steps=data_handler.inferred_steps,
-            )
-
-        callbacks.on_train_begin()
-        # data_handler._initial_epoch = (  # pylint: disable=protected-access
-        #     self._maybe_load_initial_epoch_from_ckpt(initial_epoch))
-        epoch_logs = {}
-
-        for epoch, iterator in data_handler.enumerate_epochs():
-            self.reset_metrics()
-            callbacks.on_epoch_begin(epoch)
-            logs = {}
-            with data_handler.catch_stop_iteration():
-                for step in data_handler.steps():
-                    callbacks.on_train_batch_begin(step)
-                    batch = next(iterator)
-                    # sample_weight = batch[2] if len(batch) == 3 else None
-                    x_batch, y_batch, sample_weight = unpack_x_y_sample_weight(batch)
-
-                    if drop_remaining and not data_utils.has_batch_size(
-                        batch, data_handler.batch_size
-                    ):
-                        continue
-
-                    tmp_logs = self.train_on_batch(
-                        x=x_batch,
-                        y=y_batch,
-                        sample_weight=sample_weight,
-                        class_weight=class_weight,
-                    )
-                    tmp_logs.update({"size": data_handler.batch_size})
-                    # print(epoch, step, tmp_logs["accuracy"], batch[0].shape)
-
-                    logs = tmp_logs
-                    callbacks.on_train_batch_end(step, logs)
-
-                    if self.stop_training:
-                        break
-
-            epoch_logs = copy(logs)
-            epoch_logs.update({"size": data_handler.batch_size})
-
-            # Run validation.
-            if (
-                validation_data
-                and self._should_eval(epoch, validation_freq)
-                and not self.stop_training
-            ):
-                val_x, val_y, val_sample_weight = unpack_x_y_sample_weight(
-                    validation_data
+            if validation_split:
+                # Create the validation data using the training data. Only supported for
+                # `Jax Numpy` and `NumPy` input.
+                (inputs, labels,), validation_data = data.train_validation_split(
+                    (inputs, labels),
+                    validation_split=validation_split,
+                    shuffle=False,
                 )
-                try:
-                    val_logs = self.evaluate(
-                        x=val_x,
-                        y=val_y,
-                        sample_weight=val_sample_weight,
-                        batch_size=validation_batch_size or batch_size,
-                        steps=validation_steps,
-                        callbacks=callbacks,
-                        # return_dict=True,
-                        drop_remaining=drop_remaining,
+
+            self.stop_training = False
+            data_handler = data.DataHandler(
+                x=inputs,
+                y=labels,
+                # sample_weight=sample_weight,
+                batch_size=batch_size,
+                steps_per_epoch=steps_per_epoch,
+                initial_epoch=initial_epoch,
+                epochs=epochs,
+                shuffle=shuffle,
+                # class_weight=class_weight,
+            )
+            # Container that configures and calls `tf.keras.Callback`s.
+            if not isinstance(callbacks, CallbackList):
+                callbacks = CallbackList(
+                    callbacks,
+                    add_history=True,
+                    add_progbar=verbose != 0,
+                    sigint_mode=SigIntMode.TRAIN,
+                    model=self,
+                    verbose=verbose,
+                    epochs=epochs,
+                    steps=data_handler.inferred_steps,
+                )
+
+            callbacks.on_train_begin()
+            # data_handler._initial_epoch = (  # pylint: disable=protected-access
+            #     self._maybe_load_initial_epoch_from_ckpt(initial_epoch))
+            epoch_logs = {}
+
+            for epoch, iterator in data_handler.enumerate_epochs():
+                self.reset_metrics()
+                callbacks.on_epoch_begin(epoch)
+                logs = {}
+                with data_handler.catch_stop_iteration():
+                    for step in data_handler.steps():
+                        # prev_model = self.copy()
+
+                        callbacks.on_train_batch_begin(step)
+                        batch = next(iterator)
+                        inputs, labels, _ = data.unpack_x_y_sample_weight(batch)
+
+                        if drop_remaining and not data_utils.has_batch_size(
+                            batch, data_handler.batch_size
+                        ):
+                            continue
+
+                        tmp_logs = self.train_on_batch(
+                            inputs=inputs,
+                            labels=labels,
+                        )
+                        tmp_logs.update({"size": data_handler.batch_size})
+
+                        logs = tmp_logs
+                        callbacks.on_train_batch_end(step, logs)
+
+                        if self.stop_training:
+                            break
+
+                        # utils._walk_treedef(
+                        #     jax.tree_flatten(prev_model)[1],
+                        #     jax.tree_flatten(self)[1],
+                        # )
+
+                epoch_logs = copy(logs)
+                epoch_logs.update({"size": data_handler.batch_size})
+
+                # Run validation.
+                if (
+                    validation_data
+                    and self._should_eval(epoch, validation_freq)
+                    and not self.stop_training
+                ):
+                    val_inputs, val_lables, _ = data.unpack_x_y_sample_weight(
+                        validation_data
                     )
+                    # val_inputs, val_lables = validation_data
+                    try:
+                        val_logs = self.evaluate(
+                            x=val_inputs,
+                            y=val_lables,
+                            batch_size=validation_batch_size or batch_size,
+                            steps=validation_steps,
+                            callbacks=callbacks,
+                            # return_dict=True,
+                            drop_remaining=drop_remaining,
+                        )
 
-                    val_logs = {"val_" + name: val for name, val in val_logs.items()}
-                    epoch_logs.update(val_logs)
-                except (types.MissingMethod, types.MissingModule) as e:
-                    pass
+                        val_logs = {
+                            "val_" + name: val for name, val in val_logs.items()
+                        }
+                        epoch_logs.update(val_logs)
+                    except (types.MissingMethod, types.MissingModule) as e:
+                        pass
 
-            callbacks.on_epoch_end(epoch, epoch_logs)
+                callbacks.on_epoch_end(epoch, epoch_logs)
 
-            if self.stop_training:
-                break
+                if self.stop_training:
+                    break
 
-        callbacks.on_train_end(epoch_logs)
+            callbacks.on_train_end(epoch_logs)
 
-        return self.history
+            assert self.history is not None
+
+            return self.history
 
     def evaluate(
         self,
         x: tp.Optional[tp.Any] = None,
         y: tp.Optional[tp.Any] = None,
         verbose: int = 1,
-        batch_size: tp.Optional[int] = None,
+        batch_size: tp.Optional[int] = 32,
         sample_weight: tp.Optional[np.ndarray] = None,
         class_weight: tp.Optional[tp.Mapping[str, float]] = None,
         steps: tp.Optional[int] = None,
@@ -661,82 +535,81 @@ class ModelBase(ModelCore):
         Raises:
             ValueError: in case of invalid arguments.
         """
+        if batch_size is not None:
+            batch_size = self._distributed_strategy.lift_batch_size(batch_size)
 
-        if x is None:
-            x = {}
+        with _MODEL_CONTEXT.callbacks_context(
+            self if _MODEL_CONTEXT.model is None else None
+        ):
+            if x is None:
+                x = {}
 
-        if not self.initialized:
-            self.init(
+            data_handler = data.DataHandler(
                 x=x,
                 y=y,
-                batch_size=batch_size,
-                class_weight=class_weight,
                 sample_weight=sample_weight,
-            )
-
-        data_handler = DataHandler(
-            x=x,
-            y=y,
-            sample_weight=sample_weight,
-            batch_size=batch_size,
-            steps_per_epoch=steps,
-            initial_epoch=0,
-            epochs=1,
-            shuffle=False,
-            training=False,
-        )
-
-        # Container that configures and calls `tf.keras.Callback`s.
-        if not isinstance(callbacks, CallbackList):
-            callbacks = CallbackList(
-                callbacks,
-                add_history=True,
-                add_progbar=verbose != 0,
-                model=self,
-                verbose=verbose,
+                batch_size=batch_size,
+                steps_per_epoch=steps,
+                initial_epoch=0,
                 epochs=1,
-                steps=data_handler.inferred_steps,
+                shuffle=False,
+                training=False,
             )
 
-        callbacks.on_test_begin()
+            # Container that configures and calls `tf.keras.Callback`s.
+            if not isinstance(callbacks, CallbackList):
+                callbacks = CallbackList(
+                    callbacks,
+                    add_history=True,
+                    add_progbar=verbose != 0,
+                    sigint_mode=SigIntMode.TEST,
+                    model=self,
+                    verbose=verbose,
+                    epochs=1,
+                    steps=data_handler.inferred_steps,
+                )
 
-        logs = {}
-        for _, iterator in data_handler.enumerate_epochs():
-            self.reset_metrics()
-            with data_handler.catch_stop_iteration():
-                for step in data_handler.steps():
-                    callbacks.on_test_batch_begin(step)
-                    batch = next(iterator)
-                    x_batch, y_batch, sample_weight = unpack_x_y_sample_weight(batch)
+            callbacks.on_test_begin()
 
-                    if drop_remaining and not data_utils.has_batch_size(
-                        batch, data_handler.batch_size
-                    ):
-                        continue
+            logs = {}
+            for _, iterator in data_handler.enumerate_epochs():
+                self.reset_metrics()
+                with data_handler.catch_stop_iteration():
+                    for step in data_handler.steps():
+                        callbacks.on_test_batch_begin(step)
+                        batch = next(iterator)
+                        inputs, labels, _ = data.unpack_x_y_sample_weight(batch)
 
-                    tmp_logs = self.test_on_batch(
-                        x=x_batch,
-                        y=y_batch,
-                        sample_weight=sample_weight,
-                        class_weight=None,
-                    )
-                    tmp_logs.update({"size": data_handler.batch_size})
-                    logs = tmp_logs
-                    callbacks.on_test_batch_end(step, logs)
+                        if drop_remaining and not data_utils.has_batch_size(
+                            batch, data_handler.batch_size
+                        ):
+                            continue
 
-        callbacks.on_test_end()
+                        tmp_logs = self.test_on_batch(
+                            inputs=inputs,
+                            labels=labels,
+                        )
+                        tmp_logs.update({"size": data_handler.batch_size})
+                        logs = tmp_logs
+                        callbacks.on_test_batch_end(step, logs)
 
-        return logs
+                        if self.stop_training:
+                            break
+                if self.stop_training:
+                    break
+
+            callbacks.on_test_end()
+
+            return logs
 
     def predict(
         self,
         x: tp.Optional[tp.Any] = None,
         verbose: int = 0,
-        batch_size: tp.Optional[int] = None,
+        batch_size: tp.Optional[int] = 32,
         steps: tp.Optional[int] = None,
         callbacks: tp.Union[tp.List[Callback], CallbackList, None] = None,
         drop_remaining: bool = False,
-        initialize: bool = False,
     ) -> tp.Any:
         """Generates output predictions for the input samples.
         Computation is done in batches.
@@ -780,82 +653,75 @@ class ModelBase(ModelCore):
                 or in case a stateful model receives a number of samples
                 that is not a multiple of the batch size.
         """
+        if batch_size is not None:
+            batch_size = self._distributed_strategy.lift_batch_size(batch_size)
 
-        if x is None:
-            x = {}
+        with _MODEL_CONTEXT.callbacks_context(self):
+            if x is None:
+                x = {}
 
-        if not self.initialized:
-            if initialize:
-                self.init(x=x, batch_size=batch_size)
-            else:
-                raise types.ModelNotInitialized(
-                    f"Model not initialized, please execute `model.init` or `model.init_on_batch` "
-                    "before running this method, or pass `initialize=True` to initialize with the available data "
-                    "(this might not initialize the optimizer)."
-                )
+            outputs = None
 
-        outputs = None
-
-        data_handler = DataHandler(
-            x=x,
-            batch_size=batch_size,
-            steps_per_epoch=steps,
-            initial_epoch=0,
-            epochs=1,
-            shuffle=False,
-        )
-
-        # Container that configures and calls `tf.keras.Callback`s.
-        if not isinstance(callbacks, CallbackList):
-            callbacks = CallbackList(
-                callbacks,
-                add_history=True,
-                add_progbar=verbose != 0,
-                model=self,
-                verbose=verbose,
+            data_handler = data.DataHandler(
+                x=x,
+                batch_size=batch_size,
+                steps_per_epoch=steps,
+                initial_epoch=0,
                 epochs=1,
-                steps=data_handler.inferred_steps,
+                shuffle=False,
             )
 
-        callbacks.on_predict_begin()
+            # Container that configures and calls `tf.keras.Callback`s.
+            if not isinstance(callbacks, CallbackList):
+                callbacks = CallbackList(
+                    callbacks,
+                    add_history=True,
+                    add_progbar=verbose != 0,
+                    model=self,
+                    verbose=verbose,
+                    epochs=1,
+                    steps=data_handler.inferred_steps,
+                )
 
-        for _, iterator in data_handler.enumerate_epochs():
-            self.reset_metrics()
-            with data_handler.catch_stop_iteration():
-                for step in data_handler.steps():
-                    callbacks.on_predict_batch_begin(step)
-                    batch = next(iterator)
+            callbacks.on_predict_begin()
 
-                    if drop_remaining and not data_utils.has_batch_size(
-                        batch, data_handler.batch_size
-                    ):
-                        continue
+            for _, iterator in data_handler.enumerate_epochs():
+                self.reset_metrics()
+                with data_handler.catch_stop_iteration():
+                    for step in data_handler.steps():
+                        callbacks.on_predict_batch_begin(step)
+                        batch = next(iterator)
 
-                    tmp_batch_outputs = self.predict_on_batch(x=batch[0])
-                    batch_outputs = tmp_batch_outputs
+                        if drop_remaining and not data_utils.has_batch_size(
+                            batch, data_handler.batch_size
+                        ):
+                            continue
 
-                    if outputs is None:
-                        outputs = map_structure(
-                            lambda batch_output: [batch_output], batch_outputs
-                        )
-                    else:
+                        tmp_batch_outputs = self.predict_on_batch(inputs=batch[0])
+                        batch_outputs = tmp_batch_outputs
 
-                        outputs = map_structure(
-                            map_append,
-                            outputs,
-                            batch_outputs,
+                        if outputs is None:
+                            outputs = data.map_structure(
+                                lambda batch_output: [batch_output], batch_outputs
+                            )
+                        else:
+
+                            outputs = data.map_structure(
+                                data.map_append,
+                                outputs,
+                                batch_outputs,
+                            )
+
+                        callbacks.on_predict_batch_end(
+                            step,
+                            {"outputs": batch_outputs, "size": data_handler.batch_size},
                         )
 
-                    callbacks.on_predict_batch_end(
-                        step,
-                        {"outputs": batch_outputs, "size": data_handler.batch_size},
-                    )
+            callbacks.on_predict_end()
 
-        callbacks.on_predict_end()
+            all_outputs = data.map_structure(jnp.concatenate, outputs)
 
-        all_outputs = map_structure(jnp.concatenate, outputs)
-
-        return all_outputs
+            return all_outputs
 
     def _should_eval(self, epoch, validation_freq):
         epoch = epoch + 1  # one-index the user-facing epoch.
